SystemOrganization addCategory: #'Pier-Pharo-Persistency'!

!PRKernel methodsFor: '*pier-pharo-persistency' stamp: 'NickAger 1/13/2010 08:51'!
exportOn: stream

	| reference persistencyBackup |
	
	self root enumerator with all do: [ :structure |
		Smalltalk at: #PRComponent ifPresent: [ :class |
			(structure isKindOf: class) ifTrue: [
				structure settings keys do: [ :desc |
					desc conditions: (desc conditions
						reject: [ :cond | cond key isBlock ]) ] ] ] ].
	
	reference := ReferenceStream on: stream binary.
	persistencyBackup := persistency.
	persistency := (persistencyBackup class new
			copyFrom: persistencyBackup;
			sanitize).
			
	[ [ reference nextPut: self; close ] 
		ensure: [ persistency := persistencyBackup ] ]
			on: Error
			do: [ :err | ^ self inform: 'Unable to export: ' , err messageText printString , '.' ].
! !

PRWidget subclass: #PRExportImportWidget
	instanceVariableNames: 'replace'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Pharo-Persistency'!

!PRExportImportWidget class methodsFor: 'accessing' stamp: 'lr 2/5/2009 22:52'!
groupLabel
	^ 'Pier System'! !

!PRExportImportWidget class methodsFor: 'as yet unclassified' stamp: 'lr 2/5/2009 22:44'!
isAbstract
	^ false! !

!PRExportImportWidget class methodsFor: 'accessing' stamp: 'lr 2/5/2009 22:44'!
label
	^ 'Export/Import'! !

!PRExportImportWidget methodsFor: 'actions' stamp: 'NickAger 1/13/2010 08:54'!
export: aKernel
	| stream |

	stream := RWBinaryOrTextStream on: ByteArray new.
	
	 aKernel exportOn: stream.

	self requestContext response
		document: stream contents
			mimeType: 'application/binary'
			fileName: aKernel name , '.obj';
		respond! !

!PRExportImportWidget methodsFor: 'actions' stamp: 'lr 2/6/2009 23:02'!
import: aCollection into: aKernel
	| stream reference kernel |
	stream := RWBinaryOrTextStream with: aCollection asByteArray.
	reference := ReferenceStream on: stream binary reset.
	[ kernel := reference next ]
		on: Error
		do: [ :err | ^ self inform: 'Unable to import: ' , err messageText printString , '.' ].
	(kernel isKindOf: PRKernel)
		ifFalse: [ ^ self inform: 'Invalid file format.' ].
	replace
		ifTrue: [ 
			aKernel becomeForward: kernel.
			aKernel persistency desanitize ]
		ifFalse: [
			kernel root name: (PRAddCommand new
				uniqueName: kernel root name
				in: aKernel root).
			aKernel root addChild: kernel root ].
	self sanitizeImport: kernel root.
	self context: (self context structure: kernel root)! !

!PRExportImportWidget methodsFor: 'initialization' stamp: 'lr 2/6/2009 22:39'!
initialize
	super initialize.
	replace := false! !

!PRExportImportWidget methodsFor: 'rendering' stamp: 'lr 2/5/2009 23:11'!
renderContentOn: html
	self renderExportOn: html.
	self renderImportOn: html! !

!PRExportImportWidget methodsFor: 'rendering' stamp: 'lr 2/6/2009 23:07'!
renderExportOn: html
	html heading level: 2; with: 'Export'.
	html paragraph: 'Exporting a snapshot of the kernel does not include the history or any external files. If you plan to move the site to a new location make sure to copy the complete directory structure below ' , MAExternalFileModel baseDirectory pathName printString , ' separately. Depending on the size of your site it might take a while to generate and download the snapshot.'.
	html form: [
		html paragraph: [
			html submitButton
				callback: [ self export: self context kernel ];
				with: 'Export' ] ]! !

!PRExportImportWidget methodsFor: 'rendering' stamp: 'lr 2/6/2009 23:14'!
renderImportOn: html
	html heading level: 2; with: 'Import'.
	html paragraph: 'Importing a snapshot either replaces the existing kernel or just adds all included pages to the current kernel. In either case, make sure to merge the external files into ' , MAExternalFileModel baseDirectory pathName printString , ' and to check security properties.'.
	html form multipart; with: [
		html paragraph: [
			html radioGroup: [ :group |
				html label: [
					html radioButton
						group: group;
						selected: replace not;
						callback: [ replace := false ].
					html text: 'Import Pages' ].
				html break.
				html label: [ 
					html radioButton
						group: group;
						selected: replace;
						callback: [ replace := true ].
					html text: 'Replace Kernel' ] ] ].
		html paragraph: [
			html fileUpload
				callback: [ :file | self import: file contents into: self context kernel ] ].
		html paragraph: [
			html submitButton: 'Import' ] ]! !

!PRExportImportWidget methodsFor: 'private' stamp: 'lr 2/6/2009 08:20'!
sanitizeExport: aStructure
	"Some older versions of Pier have block contexts in their conditions, get rid of these."
	
	aStructure enumerator with all do: [ :structure |
		Smalltalk at: #PRComponent ifPresent: [ :class |
			(structure isKindOf: class) ifTrue: [
				structure settings keys do: [ :desc |
					desc conditions: (desc conditions
						reject: [ :cond | cond key isBlock ]) ] ] ] ]! !

!PRExportImportWidget methodsFor: 'private' stamp: 'lr 4/1/2009 21:56'!
sanitizeImport: aStructure
	"After importing drop all the security settings as these users might not exist in the new kernel. Also refresh all the links so that they are properly printed when editing these pages."
	
	aStructure enumerator with all do: [ :structure |
		replace ifFalse: [
			Smalltalk at: #PUSecurity ifPresent: [ :class |
				structure 
					removeDecoration: class
					ifAbsent: [ ] ] ].
		structure outgoingReferences do: [ :link |
			(link isInternal and: [ link isBroken not ])
				ifTrue: [ link update ] ] ]! !

!GRPackage class methodsFor: '*pier-pharo-persistency' stamp: 'jf 11/3/2009 11:22'!
pierPharoPersistency
	^ self new
		name: 'Pier-Pharo-Persistency';
		addDependency: 'Pier-Model';
		addDependency: 'Pier-Seaside';
		addDependency: 'Pier-Tests-Model';
		yourself! !

!ImageSegment methodsFor: '*pier-pharo-persistency' stamp: 'lr 3/26/2009 23:36'!
writeKernel: aKernel on: aStream
	"Serialize aKernel on aStream so that it can be loaded into another image later. Answer a list of outpointers, these are objects that are referenced from somewhere else."

	| previousMarker |
	self
		copyFromRoots: (Array with: aKernel)
		sizeHint: 1024 *1024
		areUnique: true.
	state = #activeCopy
		ifFalse: [ self error: 'The image segment has an invalid state.' ].
	previousMarker := endMarker.
	endMarker := nil.
	aStream fileOutClass: nil andObject: self.
	endMarker := previousMarker.
	^ outPointers copy! !

PRHistoryPersistency subclass: #PRDelayedPersistency
	instanceVariableNames: 'duration process message'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Pharo-Persistency'!

!PRDelayedPersistency commentStamp: 'lr 2/1/2009 17:03' prior: 0!
A snapshot based persistency mechanism.!

!PRDelayedPersistency class methodsFor: 'descriptions' stamp: 'lr 1/29/2009 21:32'!
descriptionDuration
	^ MADurationDescription new
		comment: 'The minimum interval between two snapshots.';
		accessor: #duration;
		label: 'Duration';
		default: 1 hour;
		priority: 100;
		beRequired;
		yourself! !

!PRDelayedPersistency class methodsFor: 'descriptions' stamp: 'lr 1/29/2009 21:18'!
descriptionMessage
	^ MAStringDescription new
		accessor: #message;
		label: 'Message';
		group: 'Status';
		priority: 200;
		beReadonly;
		yourself! !

!PRDelayedPersistency class methodsFor: 'testing' stamp: 'lr 1/29/2009 21:15'!
isAbstract
	^ self name = #PRDelayedPersistency! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:16'!
duration
	^ duration! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:16'!
duration: aDuration
	"The minimum interval between two snapshots."

	duration := aDuration! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 2/20/2009 00:29'!
execute: aCommand
	self mutex critical: [
		self apply: aCommand.
		aCommand isLogged ifTrue: [
			self add: aCommand context.
			self log: aCommand context.
			self start ] ]! !

!PRDelayedPersistency methodsFor: 'testing' stamp: 'lr 1/29/2009 21:26'!
isScheduled
	^ process notNil and: [ process isTerminated not ]! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:21'!
log: aContext
	| stream |
	stream := FileDirectory default
		fileNamed: 'transactions.txt'.
	[ self log: aContext on: stream setToEnd ]
		ensure: [ stream close ]! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:21'!
log: aContext on: aStream
	aStream nextPutAll: ' == '; nextPutAll: aContext command class label; nextPutAll: ' == '; cr.
	aStream nextPutAll: ' -- context -- '; cr.
	aContext description do: [ :each |
		aStream
			nextPutAll: each label; nextPutAll: ': ';
			nextPutAll: (each toString: (aContext
				readUsing: each)); cr ].
	aStream nextPutAll: ' -- command -- '; cr.
	aContext command description do: [ :each |
		aStream
			nextPutAll: each label; nextPutAll: ': ';
			nextPutAll: (each toString: (aContext command
				readUsing: each)); cr ].
	aStream cr; cr! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 2/1/2009 16:52'!
message
	^ message! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 2/1/2009 16:52'!
message: aString
	message := aString! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 22:01'!
process
	(Delay forDuration: self duration)
		wait.
	Processor activeProcess
		priority: Processor highestPriority.
	process := nil.
	self snapshot! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 2/6/2009 22:33'!
sanitize
	super sanitize.
	self stop! !

!PRDelayedPersistency methodsFor: 'initialization' stamp: 'lr 1/29/2009 21:20'!
setKernel: aKernel
	super setKernel: aKernel.
	self stop! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:31'!
snapshot
	"Take a snapshot and perform cleanup tasks."

	self subclassResponsibility! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:30'!
start
	"Schedule the receiver to take a snapshot after the configured duration."

	self isScheduled
		ifTrue: [ ^ self ].
	process := [ self process ]
		newProcess.
	process 
		priority: Processor userBackgroundPriority;
		name: 'pier'.
	process resume! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 2/20/2009 00:31'!
stop
	"Stop the configured snapshot."

	self isScheduled
		ifTrue: [ process terminate ].
	process := nil! !

PRDelayedPersistency subclass: #PRImagePersistency
	instanceVariableNames: 'namePrefix maxImages'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Pharo-Persistency'!

!PRImagePersistency commentStamp: 'lr 2/1/2009 17:08' prior: 0!
A persistency strategy regularly saving the complete image. Do not use this strategy while developing, as this might corrupt the changes file.!

!PRImagePersistency class methodsFor: 'accessing-descriptions' stamp: 'lr 1/23/2009 14:02'!
descriptionMaxImages
	^ MANumberDescription new
		comment: 'The maximum number of backup copies to keep';
		accessor: #maxImages;
		label: 'Backup Images';
		priority: 110;
		beRequired;
		default: 3;
		beInteger;
		yourself! !

!PRImagePersistency class methodsFor: 'accessing-descriptions' stamp: 'lr 1/23/2009 14:03'!
descriptionNamePrefix
	^ MAStringDescription new
		comment: 'Defines the prefix for image-names.';
		default: 'pier_backup_';
		accessor: #namePrefix;
		label: 'Backup Prefix';
		priority: 120;
		beRequired;
		yourself! !

!PRImagePersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:31'!
cleanupBackupImages
	| images |
	maxImages isNil ifTrue: [ ^ self ].
	images := FileDirectory default fileNames select: [ :each | 
		(each endsWith: FileDirectory imageSuffix) 
			and: [ each beginsWith: namePrefix ] ].
	images size <= maxImages ifTrue: [ ^ self ].
	images := (images asSortedCollection allButLast: maxImages)
		collect: [ :each | each copyUpToLast: $. ].
	images do: [ :each |
		FileDirectory default 
			deleteFileNamed: (each , '.' , FileDirectory imageSuffix)
			ifAbsent: [ ].
		FileDirectory default 
			deleteFileNamed: (each , '.' , FileDirectory changeSuffix)
			ifAbsent: [ ] ]! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 6/22/2008 17:54'!
maxImages
	^ maxImages! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:27'!
maxImages: anInteger
	"The maximum number of backup copies to keep."

	maxImages := anInteger! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 6/22/2008 18:06'!
namePrefix
	^ namePrefix! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:27'!
namePrefix: aString
	"Defines the prefix for backup copies."

	namePrefix := aString! !

!PRImagePersistency methodsFor: 'private' stamp: 'lr 4/20/2009 22:46'!
saveImageAndBackupAs: aString
	| directory image imageName changeName |
	directory := FileDirectory default.
	image := SmalltalkImage current.
	image closeSourceFiles.
	imageName := aString , '.' , FileDirectory imageSuffix.
	changeName := aString , '.' , FileDirectory changeSuffix.
	((directory fileExists: imageName) not and: [ (directory fileExists: changeName) not ]) ifTrue: [
		directory
			copyFileWithoutOverwriteConfirmationNamed: image imageName 
			toFileNamed: (directory fullPathFor: imageName).
		directory
			copyFileWithoutOverwriteConfirmationNamed: image changesName 
			toFileNamed: (directory fullPathFor: changeName).	
		Smalltalk setMacFileInfoOn: (directory fullPathFor: changeName) ].
	image openSourceFiles; saveImageSegments; snapshot: true andQuit: false! !

!PRImagePersistency methodsFor: 'actions' stamp: 'jf 11/23/2009 20:58'!
snapshot
	| filename |
	filename := namePrefix , DateAndTime now asSortableString.
	[ self saveImageAndBackupAs: filename ]
		valueUnpreemptively.
	self message: 'Last snapshot ' , filename printString , ' at ' , TimeStamp now greaseString.
	self cleanupBackupImages! !

PRDelayedPersistency subclass: #PRSegmentPersistency
	instanceVariableNames: 'namePrefix'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Pharo-Persistency'!

!PRSegmentPersistency commentStamp: 'lr 2/1/2009 17:05' prior: 0!
A persistency strategy regularly saving an image segment of the kernel. This requires the OS process plugin to work.!

!PRSegmentPersistency class methodsFor: 'accessing-descriptions' stamp: 'lr 2/5/2009 16:35'!
descriptionNamePrefix
	^ MAStringDescription new
		comment: 'Defines the prefix for snapshot files.';
		label: 'Snapshot Prefix';
		accessor: #namePrefix;
		default: 'snapshot_';
		priority: 120;
		beRequired;
		yourself! !

!PRSegmentPersistency class methodsFor: 'testing' stamp: 'lr 1/30/2009 16:55'!
isAbstract
	^ true! !

!PRSegmentPersistency methodsFor: 'private' stamp: 'lr 4/6/2009 23:58'!
cleanup
	"Execute all the registered cleanup code. Do not execute this code within an image you intend to continue using. This is a non-reversible operation and pretty much destroys your complete environment."

	(Pragma allNamed: #cleanup from: self class to: PRSegmentPersistency)
		do: [ :pragma | self perform: pragma selector ].
	Smalltalk garbageCollect! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:04'!
cleanupKom
	"Stop all the server adaptors, they just get into our way."
	<cleanup>
	
	WAKom withAllSubclassesDo: [ :each | each shutDown ]! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'jf 9/15/2009 18:03'!
cleanupMagritte
	"Clear the magritte cache that shares common description objects with the kernel."
	<cleanup>
	
	MADescriptionBuilder default flush! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:21'!
cleanupMonticello
	"Clear monticello caches, they might reference some of our blocks in obscure ways."
	<cleanup>
	
	MCMethodDefinition shutDown.
	MCFileBasedRepository flushAllCaches! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:04'!
cleanupPier
	"Clear all registered kernels. We don't want references to our model from outside."
	<cleanup>
	
	PRKernel instVarNamed: 'Instances' put: nil! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:13'!
cleanupProcesses
	"Get rid of pocesses that potentialy still hang somewhere and reference our model. Take care not to kill any of the vital processes."
	<cleanup>

	| protected |
	protected := IdentitySet new.
	protected
		add: Project uiProcess;
		add: Processor activeProcess;
		add: Sensor eventTicklerProcess;
		add: Processor backgroundProcess;
		add: Sensor interruptWatcherProcess;
		add: Smalltalk lowSpaceWatcherProcess;
		add: WeakArray runningFinalizationProcess.
	Process allInstances do: [ :each |
		(each isTerminated or: [ Processor highestPriority = each priority or: [ protected includes: each ] ])
			ifFalse: [ each terminate ] ]
	.
	! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'NickAger 1/8/2010 19:27'!
cleanupSeaside

	<cleanup>

    WAAdmin clearAll! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:18'!
cleanupWorld
	"Close all windows, just in case we have some workspaces or pier browsers open that reference our model."
	<cleanup>

	World submorphs do: [ :morph | morph delete ].
	World releaseCachedState! !

!PRSegmentPersistency methodsFor: 'accessing' stamp: 'lr 2/5/2009 16:35'!
directory
	^ (FileDirectory default directoryNamed: 'snapshots') assureExistence! !

!PRSegmentPersistency methodsFor: 'accessing' stamp: 'lr 2/5/2009 16:30'!
namePrefix
	^ namePrefix! !

!PRSegmentPersistency methodsFor: 'accessing' stamp: 'lr 2/5/2009 16:30'!
namePrefix: aString
	namePrefix := aString! !

!PRSegmentPersistency methodsFor: 'private' stamp: 'lr 1/30/2009 08:34'!
serialize: aString
	| stream pointers |
	stream := self directory forceNewFileNamed: aString.
	pointers := [ [ ImageSegment new writeKernel: self kernel on: stream ]
		ensure: [ stream close ] ] on: Exception do: [ :error | nil ].
	(pointers isNil or: [ pointers anySatisfy: [ :each | (each isLiteral or: [ each isVariableBinding ]) not ] ]) 
		ifTrue: [ self directory deleteFileNamed: aString ifAbsent: [ ] ]! !

!PRSegmentPersistency methodsFor: 'actions' stamp: 'jf 11/3/2009 11:08'!
snapshot
	| os filename |
	(os := Smalltalk classNamed: #OSProcess) isNil
		ifTrue: [ ^ self message: 'OSProcess not installed.' ].
	os thisOSProcess pid
		ifTrue: [ ^ self message: 'OSProcess does not support this platform.' ].
	filename := self namePrefix , DateAndTime now asSortableString , '.obj'.
	os thisOSProcess forkHeadlessSqueakAndDo: [
		[ self cleanup; serialize: filename ]
			ensure: [ Smalltalk quitPrimitive ] ].
	self message: 'Last snapshot ' , filename printString , ' at ' , GRPlatform current magritteTimeStampClass now greaseString! !

!PRHistoryPersistency methodsFor: '*pier-pharo-persistency' stamp: 'lr 2/6/2009 22:31'!
sanitize
	super sanitize.
	history := OrderedCollection new! !

PRPersistencyTest subclass: #PRImagePersistencyTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Pharo-Persistency'!

!PRImagePersistencyTest class methodsFor: 'testing' stamp: 'lr 3/30/2007 22:22'!
isAbstract
	^ false! !

!PRImagePersistencyTest methodsFor: 'private' stamp: 'lr 3/30/2007 22:22'!
actualClass
	^ PRImagePersistency! !

!PRPersistency methodsFor: '*pier-pharo-persistency' stamp: 'lr 2/6/2009 22:33'!
desanitize
	"Sanitizes the receiver to be serializeable."
	
	mutex := self defaultMutex! !

!PRPersistency methodsFor: '*pier-pharo-persistency' stamp: 'lr 2/6/2009 22:32'!
sanitize
	"Sanitizes the receiver to be serializeable."
	
	mutex := nil! !

!DateAndTime methodsFor: '*pier-pharo-persistency' stamp: 'lr 9/23/2008 08:13'!
asSortableString
	"A small interpreter to get a date exactly the way required."

	| code |
	code := #( year 4 $- monthIndex 2 $- dayOfMonth 2 $_ hour 2 $- minute 2 $- second 2 ) readStream.
	^ String streamContents: [ :stream |
		[ code atEnd ] whileFalse: [
			stream nextPutAll: ((self perform: code next)
				printPaddedWith: $0 to: code next).
			code next
				ifNotNilDo: [ :separator | stream nextPut: separator ] ] ]! !
